@@ -23,15 +23,17 @@ from collections import defaultdict
 import json
 import logging
 import os
-import shutil
 import sys
-from typing import Any, Callable, DefaultDict, Dict, Iterator, List, Tuple
+from typing import Any, DefaultDict, Dict, Iterator, List, Tuple
+import zipfile
 from schema import (Optional, Or, Schema, SchemaError, SchemaMissingKeyError,
                     SchemaForbiddenKeyError, SchemaUnexpectedTypeError)
 import yaml
 
 import boto3
 
+HELPERS_LOCATION = 'global_helpers'
+
 
 class TestCase():
 
@@ -58,6 +60,7 @@ GLOBAL_SCHEMA = Schema(
     {
         'AnalysisType': Or("global"),
         'Filename': str,
+        'GlobalID': str,
         Optional('Description'): str,
         Optional('Tags'): [str],
     },
@@ -222,10 +225,15 @@ def zip_analysis(args: argparse.Namespace) -> Tuple[int, str]:
     # The colon character is not valid in filenames.
     current_time = datetime.now().isoformat(timespec='seconds').replace(
         ':', '-')
-    filename = 'panther-analysis'
-    return 0, shutil.make_archive(
-        os.path.join(args.out, '{}-{}'.format(filename, current_time)), 'zip',
-        args.path)
+    filename = 'panther-analysis-{}.zip'.format(current_time)
+    with zipfile.ZipFile(filename, 'w', zipfile.ZIP_DEFLATED) as zip_out:
+        analysis = filter_analysis(list(load_analysis_specs(args.path)),
+                                   args.filter)
+        for analysis_spec_filename, dir_name, analysis_spec in analysis:
+            zip_out.write(analysis_spec_filename)
+            zip_out.write(os.path.join(dir_name, analysis_spec['Filename']))
+
+    return 0, filename
 
 
 def upload_analysis(args: argparse.Namespace) -> Tuple[int, str]:
@@ -299,8 +307,13 @@ def test_analysis(args: argparse.Namespace) -> Tuple[int, list]:
     logging.info('Testing analysis packs in %s\n', args.path)
 
     # First classify each file
-    specs = list(load_analysis_specs(args.path))
-    global_analysis, analysis, invalid_specs = classify_analysis(specs)
+    global_analysis, analysis, invalid_specs = classify_analysis(
+        list(load_analysis_specs(args.path)) +
+        list(load_analysis_specs(HELPERS_LOCATION)))
+
+    # Apply the filters as needed
+    global_analysis = filter_analysis(global_analysis, args.filter)
+    analysis = filter_analysis(analysis, args.filter)
 
     # First import the globals
     for analysis_spec_filename, dir_name, analysis_spec in global_analysis:
@@ -310,7 +323,7 @@ def test_analysis(args: argparse.Namespace) -> Tuple[int, list]:
         if load_err:
             invalid_specs.append((analysis_spec_filename, load_err))
             break
-        sys.modules['panther'] = module
+        sys.modules[analysis_spec['GlobalID']] = module
 
     # Next import each policy or rule and run its tests
     for analysis_spec_filename, dir_name, analysis_spec in analysis:
@@ -333,10 +346,17 @@ def test_analysis(args: argparse.Namespace) -> Tuple[int, list]:
             continue
 
         tests.append(analysis_id)
+        analysis_funcs = {}
         if analysis_spec['AnalysisType'] == 'policy':
-            failed_tests = run_tests(analysis_spec, module.policy, failed_tests)
+            analysis_funcs['run'] = module.policy
         elif analysis_spec['AnalysisType'] == 'rule':
-            failed_tests = run_tests(analysis_spec, module.rule, failed_tests)
+            analysis_funcs['run'] = module.rule
+            if 'dedup' in dir(module):
+                analysis_funcs['dedup'] = module.dedup
+            if 'title' in dir(module):
+                analysis_funcs['title'] = module.title
+
+        failed_tests = run_tests(analysis_spec, analysis_funcs, failed_tests)
         print('')
 
     for analysis_id in failed_tests:
@@ -349,6 +369,20 @@ def test_analysis(args: argparse.Namespace) -> Tuple[int, list]:
     return int(bool(failed_tests or invalid_specs)), invalid_specs
 
 
+def filter_analysis(analysis: List[Any], filters: Dict[str, List]) -> List[Any]:
+    if filters is None:
+        return analysis
+
+    filtered_analysis = []
+    for file_name, dir_name, analysis_spec in analysis:
+        if all(
+                analysis_spec.get(key, "") in values
+                for key, values in filters.items()):
+            filtered_analysis.append((file_name, dir_name, analysis_spec))
+
+    return filtered_analysis
+
+
 def classify_analysis(
     specs: List[Tuple[str, str, Any]]
 ) -> Tuple[List[Any], List[Any], List[Any]]:
@@ -380,7 +414,7 @@ def classify_analysis(
     return (global_analysis, analysis, invalid_specs)
 
 
-def run_tests(analysis: Dict[str, Any], run_func: Callable[[TestCase], bool],
+def run_tests(analysis: Dict[str, Any], analysis_funcs: Dict[str, Any],
               failed_tests: DefaultDict[str, list]) -> DefaultDict[str, list]:
 
     # First check if any tests exist, so we can print a helpful message if not
@@ -394,7 +428,7 @@ def run_tests(analysis: Dict[str, Any], run_func: Callable[[TestCase], bool],
             test_case = TestCase(
                 unit_test.get('Resource') or unit_test['Log'],
                 unit_test.get('ResourceType') or unit_test['LogType'])
-            result = run_func(test_case)
+            result = analysis_funcs['run'](test_case)
         except KeyError as err:
             print("KeyError: {0}".format(err))
             continue
@@ -404,6 +438,10 @@ def run_tests(analysis: Dict[str, Any], run_func: Callable[[TestCase], bool],
             failed_tests[analysis.get('PolicyID') or
                          analysis['RuleID']].append(unit_test['Name'])
         print('\t[{}] {}'.format(test_result, unit_test['Name']))
+        if analysis_funcs.get('title'):
+            print('\t\t[Title] {}'.format(analysis_funcs['title'](test_case)))
+        if analysis_funcs.get('dedup'):
+            print('\t\t[Dedup] {}'.format(analysis_funcs['dedup'](test_case)))
 
     return failed_tests
 
@@ -415,7 +453,7 @@ def setup_parser() -> argparse.ArgumentParser:
         prog='panther_analysis_tool')
     parser.add_argument('--version',
                         action='version',
-                        version='panther_analysis_tool 0.2.1')
+                        version='panther_analysis_tool 0.3.0')
     subparsers = parser.add_subparsers()
 
     test_parser = subparsers.add_parser(
@@ -426,6 +464,10 @@ def setup_parser() -> argparse.ArgumentParser:
         type=str,
         help='The relative path to Panther policies and rules.',
         required=True)
+    test_parser.add_argument('--filter',
+                             required=False,
+                             metavar="KEY=VALUE",
+                             nargs='+')
     test_parser.set_defaults(func=test_analysis)
 
     zip_parser = subparsers.add_parser(
@@ -443,6 +485,10 @@ def setup_parser() -> argparse.ArgumentParser:
         type=str,
         help='The path to write zipped policies and rules to.',
         required=True)
+    zip_parser.add_argument('--filter',
+                            required=False,
+                            metavar="KEY=VALUE",
+                            nargs='+')
     zip_parser.set_defaults(func=zip_analysis)
 
     upload_parser = subparsers.add_parser(
@@ -460,11 +506,34 @@ def setup_parser() -> argparse.ArgumentParser:
         help=
         'The location to store a local copy of the packaged policies and rules.',
         required=False)
+    upload_parser.add_argument('--filter',
+                               required=False,
+                               metavar="KEY=VALUE",
+                               nargs='+')
     upload_parser.set_defaults(func=upload_analysis)
 
     return parser
 
 
+# Parses the filters, expects a list of strings
+def parse_filter(filters: List[str]) -> Dict[str, Any]:
+    parsed_filters = {}
+    for filt in filters:
+        split = filt.split('=')
+        if len(split) != 2 or split[0] == '' or split[1] == '':
+            logging.warning('Filter %s is not in format KEY=VALUE, skipping',
+                            filt)
+            continue
+        key = split[0]
+        if key not in list(GLOBAL_SCHEMA.schema.keys()) + list(
+                POLICY_SCHEMA.schema.keys()) + list(RULE_SCHEMA.schema.keys()):
+            logging.warning(
+                'Filter key %s is not a valid filter field, skipping', key)
+            continue
+        parsed_filters[key] = split[1].split(',')
+    return parsed_filters
+
+
 def run() -> None:
     logging.basicConfig(format='[%(levelname)s]: %(message)s',
                         level=logging.INFO)
@@ -472,6 +541,8 @@ def run() -> None:
     parser = setup_parser()
     args = parser.parse_args()
     try:
+        if args.filter is not None:
+            args.filter = parse_filter(args.filter)
         return_code, out = args.func(args)
     except AttributeError:
         parser.print_help()
